use crate::{executor::{Executor, MemoryData, MemoryRustFunction}, parser::ASTPart};

fn to_printed(from: &ASTPart) -> String {
    match from {
        ASTPart::String(s) => s.value.clone(),
        ASTPart::Number(n) => n.value.to_string(),
        ASTPart::Boolean(b) => b.value.to_string(),
        ASTPart::Null(_) => String::from("null"),
        _ => String::new()
    }
}

fn ugass(_exec: &mut Executor, args: Vec<ASTPart>) -> ASTPart {
    let mut out = String::new();
    for arg in &args {
        out.push_str(&to_printed(arg));
        out.push_str("  ");
    }
    println!("{}", out);
    return ASTPart::NOOP;
}

fn print_memory(exec: &mut Executor, _args: Vec<ASTPart>) -> ASTPart {
    println!("Memory dump: {:?}", exec.memory);
    return ASTPart::NOOP;
}

pub fn generate() -> Vec<MemoryData> {
    let mut out: Vec<MemoryData> = vec![];
    out.push(MemoryData::RustFunction(MemoryRustFunction { key: String::from("ugass"), func: ugass }));
    out.push(MemoryData::RustFunction(MemoryRustFunction { key: String::from("printMemory"), func: print_memory }));
    return out;
}